

'   /**
'   ---------------------------------------------
'   [- (A) Contenido -]
'   ---------------------------------------------
'   [- Versin       -] = 23.12.30.1515.
'   [- Nombre Clase  -] = clsLZW.
'   -------------
'   [- Descripcin   -] = compresin de datos LZW.
'   -------------
'   [- Observaciones -] = 
'
'    ---------------------
'      Fundamentos de LZW
'      El enfoque original de Lempel Ziv para la compresin de datos se public por primera vez en 1977,
'      seguido de un enfoque alternativo en 1978. En 1978 se publicaron los refinamientos
'      de Terry Welch al algoritmo en 1984. El algoritmo es sorprendentemente simple.
'      En pocas palabras, la compresin LZW reemplaza a las cadenas de caracteres con cdigos nicos.
'      No hace ningn anlisis del texto entrante. En cambio, simplemente agrega cada nueva
'      cadena de caracteres que ve a una tabla de cadenas. La compresin se produce cuando Se genera
'      un solo cdigo en lugar de una cadena de caracteres.
'      El cdigo que genera el algoritmo LZW puede ser de cualquier longitud arbitraria, pero debe tener
'      ms bits en l que un solo carcter. Los primeros 256 cdigos (cuando se utilizan caracteres de ocho bits)
'      son Predeterminado asignado al juego de caracteres estndar. Los cdigos restantes se asignan
'      a las cadenas como El algoritmo contina.
'      El programa de ejemplo se ejecuta como se muestra con cdigos de 12 bits. Esto significa que Los
'      cdigos 0-255 se refieren a bytes individuales, mientras que los cdigos 256-4095 se refieren a subcadenas.
'
'
'    ---------------------
'      Compresin
'      El algoritmo de compresin LZW en su forma ms simple se muestra en la Figura 1.
'      Un examen rpido de el algoritmo muestra que LZW siempre est tratando de generar cdigos
'      para cadenas que ya se conocen. Y cada vez que se genera un nuevo cdigo, se agrega una nueva cadena a la tabla de cadenas.
'
'      Rutina LZW_COMPRESS
'      STRING = get input character
'      WHILE there are still input characters DO
'        CHARACTER = get input character
'        IF STRING+CHARACTER is in the string table then
'            STRING = STRING+character
'        ELSE
'            output the code for STRING
'            add STRING+CHARACTER to the string table
'            STRING = CHARACTER
'        END of IF
'      END of WHILE
'      output the code for STRING
'
'    ---------------------
'      Descompresin
'      El algoritmo complementario para la compresin es el algoritmo de descompresin.
'      Tiene que ser capaz de Tome el flujo de cdigos de salida del algoritmo de compresin
'      y utilcelos para recrear exactamente el flujo de entrada.
'      Una de las razones de la eficiencia del algoritmo LZW es que no necesita Pase la tabla de cadenas
'      al cdigo de descompresin. La mesa se puede construir exactamente como estaba durante compresin,
'      utilizando el flujo de entrada como datos. Esto es posible porque el algoritmo de compresin
'      siempre genera los componentes STRING y CHARACTER de un cdigo antes de usarlo en la salida corriente.
'      Esto significa que los datos comprimidos no estn cargados con el transporte de una cadena grande tabla de traduccin.
'
'      Rutina LZW_DECOMPRESS
'      Read OLD_CODE
'      output OLD_CODE
'      WHILE there are still input characters DO
'        Read NEW_CODE
'        STRING = get translation of NEW_CODE
'        output STRING
'        CHARACTER = first character in STRING
'        add OLD_CODE + CHARACTER to the translation table
'        OLD_CODE = NEW_CODE
'      END of WHILE.
'
'   -------------
'   [- Bibliografa  -] = 
'      https://marknelson.us/
'      https://marknelson.us/posts/1989/10/01/lzw-data-compression
'      https://marknelson.us/posts/2011/11/08/lzw-revisited
'   -------------
'   [- CodigoEjemplo -] = cdigo ejemplo.
'   ---------------------------------------------
'   [- (B) Informacin bsica -]
'   ---------------------------------------------
'   [- Module Name   -] = clsLZW.vb.
'   [- Type          -] = vb/file.
'   [- Size          -] = 11,09 KB - [11359 Bytes].
'   [- Lenguaje      -] = spa [es-ES] Espaol (Espaa).
'   -------------
'   [- Fecha de creacin       -] = 2006-08-24T16:12:54.
'   [- ltima modificacin     -] = 2023-12-30T15:14:07.
'   -------------
'   [- Net Framework Release   -] = Release: 4.6.1 o posterior.
'   [- Net Framework Version   -] = Versin: 4.0.30319.42000.
'   ---------------------------------------------
'   [- (C) Propiedad intelectual -]
'   ---------------------------------------------
'   [- Autor Nombre    -] = Marcos Nelson.
'   [- Autor CorreoE   -] = https://marknelson.us/.
'   ---------------------------------------------
'   [- (D) Revisiones -]
'   ---------------------------------------------
'   [- Historial de revisiones -] = 
'      - 2006/08/29  - Creacin
'      - 2023/12/30  - Actualizacin y documentacin.
'   ---------------------------------------------
'   [- /Eof -]
'   */


Option Explicit On
Option Strict On
Option Infer Off
Option Compare Binary






''' <summary>
'''    compresi&#243;n de datos LZW
''' </summary>
''' <remarks>
'''    <dcterms.abstract>
'''       <para> Fundamentos de LZW </para> 
'''       <para> 
'''           El enfoque original de Lempel Ziv para la compresin de datos se public por primera vez en 1977,
'''           seguido de un enfoque alternativo en 1978. En 1978 se publicaron los refinamientos
'''           de Terry Welch al algoritmo en 1984. 
'''       </para> 
'''       <para> 
'''            El algoritmo es sorprendentemente simple.En pocas palabras,
'''            la compresin LZW reemplaza a las cadenas de caracteres con cdigos nicos.
'''            No hace ningn anlisis del texto entrante. En cambio, simplemente agrega cada nueva
'''            cadena de caracteres que ve a una tabla de cadenas. La compresin se produce cuando Se genera
'''            un solo cdigo en lugar de una cadena de caracteres.
'''       </para> 
'''       <para> 
'''           El cdigo que genera el algoritmo LZW puede ser de cualquier longitud arbitraria, pero debe tener
'''           ms bits en l que un solo carcter. Los primeros 256 cdigos (cuando se utilizan caracteres de ocho bits)
'''           son Predeterminado asignado al juego de caracteres estndar. Los cdigos restantes se asignan
'''           a las cadenas como El algoritmo contina.
'''       </para> 
'''    </dcterms.abstract>
'''    <dcterms.bibliographicCitation>
'''       <para> https://marknelson.us/ </para> 
'''       <para> https://marknelson.us/posts/1989/10/01/lzw-data-compression </para> 
'''       <para> https://marknelson.us/posts/2011/11/08/lzw-revisited </para> 
'''    </dcterms.bibliographicCitation>
'''    <example>
'''       Un Ejemplo con la forma de uso de la clase 
'''       <code>
'''          <![CDATA[
'''             codigo ejemplo
'''          ]]>
'''       </code>
'''    </example>
'''    <dcterms.created>
'''       <scheme>dcterms.ISO8601</scheme>
'''       <content>2006-08-24</content>
'''    </dcterms.created>
'''    <dcterms.modified>
'''       <scheme>dcterms.ISO8601</scheme>
'''       <content>2023-12-30T16:38:54</content>
'''    </dcterms.modified>
'''    <dc.creator>
'''       <scheme>dcterms.DCSV</scheme>
'''       <content>Marcos Nelson</content>
'''       <content>https://marknelson.us/</content>
'''    </dc.creator>
'''    <dc.publisher>
'''       <scheme>dcterms.DCSV</scheme>
'''       <content>Joaquin Medina Serrano</content>
'''       <content>mailto:gmjms32@gmail.com</content>
''' </dc.publisher>
''' </remarks>
Public Class clsLZW

    Private BITS As Integer = 14
    Private HASHING_SHIFT As Integer = 4
    Private MAX_VALUE As Integer = (1 << BITS) - 1
    Private MAX_CODE As Integer = MAX_VALUE - 1


    Private Const TABLE_SIZE As Integer = 18041 ' 14 bits
    Private EOF As Integer = -1

    Public brInput As IO.BinaryReader = Nothing
    Public bwOutput As IO.BinaryWriter = Nothing

    Private iaCode_Value(TABLE_SIZE) As Integer
    Private iaPrefix_Code(TABLE_SIZE) As Integer
    Private baAppend_Character(TABLE_SIZE) As Byte




    ''' <summary>
    '''  Esta es la rutina de compresin
    ''' </summary>
    Public Sub compress()
        Dim iNextCode As Integer = 0
        Dim iCharacter As Integer = 0
        Dim iStringCode As Integer = 0
        Dim iIndex As Integer = 0


        ' El siguiente cdigo es el siguiente cdigo de cadena disponible
        iNextCode = 256


        ' Limpiamos la tabla de cadenas antes de comenzar
        For i As Integer = 0 To TABLE_SIZE - 1
            iaCode_Value(i) = -1
        Next i

        'Obtiene el primer carcter. Suponiendo que sea 0 - 255
        'Por lo tanto, solo es vlido para archivos de texto ASCII 
        iStringCode = ReadByte()


        '** Este es el bucle principal donde sucede todo. Este bucle se ejecuta hasta que toda
        '** la entrada se ha agotado. Tenga en cuenta que deja de agregar cdigos a la
        '** tabla despus de que se hayan definido todos los cdigos posibles.
        iCharacter = ReadByte()
        While iCharacter <> -1
            iIndex = find_match(iStringCode, iCharacter)    ' Ver si la cadena est en
            If (iaCode_Value(iIndex) <> -1) Then            ' la mesa. Si esto es,
                iStringCode = iaCode_Value(iIndex)          ' obtiene el valor del cdigo. Si
            Else                                            ' la cadena no est en la tabla, intenta agregarla.
                If (iNextCode <= MAX_CODE) Then
                    iaCode_Value(iIndex) = iNextCode
                    iNextCode += 1
                    iaPrefix_Code(iIndex) = iStringCode
                    baAppend_Character(iIndex) = CByte(iCharacter)
                End If
                output_code(iStringCode)    ' Cuando se encuentra una cadena 
                iStringCode = iCharacter    ' que no est en la tabla 
            End If                          ' Despus de agregar el nuevo 
            iCharacter = ReadByte()
        End While

        'Fin del bucle principal.
        output_code(iStringCode)  ' Muestra el ltimo cdigo              
        output_code(MAX_VALUE)    ' Muestra el final del cdigo del buffer   
        output_code(0)            ' Este cdigo vaca el buffer de salida
    End Sub



    ' Esta es la rutina de hash.
    ' Intenta encontrar una coincidencia para la cadena prefijo+carcter en la tabla de cadenas.
    ' Si lo encuentra, se devuelve el iIndex (indice) que ocupa.
    ' Si la cadena no se encuentra, se devuelve el primer iIndex (indice) disponible en la tabla de cadenas


    ''' <summary>
    '''   Esta es la rutina de hash.
    ''' </summary>
    ''' <param name="iHashPrefix"></param>
    ''' <param name="iHashCharacter"></param>
    ''' <returns>
    ''' Intenta encontrar una coincidencia para la cadena prefijo+carcter en la tabla de cadenas.
    ''' Si lo encuentra, se devuelve el iIndex (indice) que ocupa.
    ''' Si la cadena no se encuentra, se devuelve el primer iIndex (indice) disponible en la tabla de cadenas
    ''' </returns>
    Private Function find_match(ByVal iHashPrefix As Integer, ByVal iHashCharacter As Integer) As Integer
        Dim iIndex As Integer = 0
        Dim iOffset As Integer = 0

        iIndex = CInt((iHashCharacter << HASHING_SHIFT) Xor iHashPrefix)

        If (iIndex = 0) Then
            iOffset = 1
        Else
            iOffset = TABLE_SIZE - iIndex
        End If

        While (True)
            If iaCode_Value(iIndex) = -1 Then
                Return iIndex
            End If
            If (iaPrefix_Code(iIndex) = iHashPrefix) And (baAppend_Character(iIndex) = iHashCharacter) Then
                Return iIndex
            End If
            iIndex -= iOffset
            If (iIndex < 0) Then
                iIndex += TABLE_SIZE
            End If
        End While
    End Function




    ''' <summary>
    ''' La siguiente rutina se utiliza para generar cdigos de longitud variable 
    ''' Est escrito estrictamente para mayor claridad y no es particularmente eficiente.
    ''' </summary>
    ''' <param name="code"></param>
    Private Sub output_code(ByVal code As Integer)
        Static output_bit_count As Integer = 0
        Static output_bit_buffer As Long = 0

        output_bit_buffer = output_bit_buffer Or (code << (32 - BITS - output_bit_count))
        output_bit_count += BITS

        While output_bit_count >= 8
            WriteByte(CByte((output_bit_buffer >> 24) And 255))
            output_bit_buffer <<= 8
            output_bit_count -= 8
        End While
    End Sub





    ''' <summary>
    ''' Esta es la rutina de expansin. Toma un archivo en formato LZW y lo expande.
    ''' a un archivo bwOutput. El cdigo aqu debera coincidir bastante con
    ''' el algoritmo en el artculo adjunto.
    ''' </summary>
    Public Sub expand()
        Dim baDecode_Stack(TABLE_SIZE) As Byte
        Dim iNextCode As Integer
        Dim iNewCode As Integer
        Dim iOldCode As Integer
        Dim bCharacter As Byte
        Dim iCurrCode As Integer
        Dim i As Integer

        'Este es el siguiente cdigo disponible para definir.
        iNextCode = 256

        ' Lee el primer cdigo,
        ' inicializar la variable de carcter y
        ' enva el primer cdigo al archivo de salida.

        iOldCode = input_code()
        bCharacter = CType(iOldCode, Byte)
        WriteByte(CByte(iOldCode))


        ' Este es el principal circuito de expansin. Lee caracteres del archivo LZW 
        ' hasta que ve el cdigo especial utilizado para indicar el final de los datos.
        iNewCode = input_code()
        While (iNewCode <> MAX_VALUE)
            If iNewCode >= iNextCode Then

                ' Este cdigo busca el carcter especial STRING+CHARACTER+STRING+CHARACTER+STRING
                ' que genera un cdigo indefinido. Lo maneja decodificando
                ' el ltimo cdigo y agregando un solo carcter al final de la cadena de decodificacin.

                baDecode_Stack(0) = bCharacter
                i = 1
                iCurrCode = iOldCode
            Else
                ' De lo contrario, hacemos una decodificacin directa del nuevo cdigo.
                i = 0
                iCurrCode = iNewCode
            End If
            While iCurrCode > 255

                ' Esta rutina simplemente decodifica una cadena de la tabla de cadenas,
                ' almacenando en un bfer. El bfer se puede generar entonces en orden inverso mediante
                ' el programa de expansin.

                baDecode_Stack(i) = baAppend_Character(iCurrCode)
                i = i + 1
                If i >= MAX_CODE Then
                    Throw New ApplicationException("Error fatal durante la expansin de iCurrCode.")
                End If
                iCurrCode = iaPrefix_Code(iCurrCode)
            End While
            baDecode_Stack(i) = CType(iCurrCode, Byte)
            bCharacter = baDecode_Stack(i)

            'Ahora generamos la cadena decodificada en orden inverso.
            While i >= 0
                WriteByte(baDecode_Stack(i))
                i = i - 1
            End While

            ' Finalmente, si es posible, agregue un nuevo cdigo a la tabla de cadenas.
            If (iNextCode <= MAX_CODE) Then
                iaPrefix_Code(iNextCode) = iOldCode
                baAppend_Character(iNextCode) = bCharacter
                iNextCode += 1
            End If
            iOldCode = iNewCode
            iNewCode = input_code()
        End While
    End Sub



    ''' <summary>
    '''   La siguiente rutina se utiliza para ingresar un cdigos de longitud variable
    '''   Est escrito estrictamente para mayor claridad y no es particularmente eficiente.
    ''' </summary>
    ''' <returns></returns>
    Private Function input_code() As Integer
        Dim return_value As Long
        Static input_bit_count As Integer = 0
        Static input_bit_buffer As Long = 0
        Static Mask32 As Long = CLng(2 ^ 32) - 1

        While input_bit_count <= 24
            input_bit_buffer = (input_bit_buffer Or
                ReadByte() << (24 - input_bit_count)) And Mask32
            input_bit_count += 8
        End While
        return_value = (input_bit_buffer >> 32 - BITS) And Mask32
        input_bit_buffer = (input_bit_buffer << BITS) And Mask32
        input_bit_count -= BITS
        Return CInt(return_value)
    End Function

    Private Sub WriteByte(ByVal b As Byte)
        bwOutput.Write(b)
    End Sub

    Private Function ReadByte() As Integer
        Dim ba(1) As Byte
        Dim iResult As Integer
        iResult = brInput.Read(ba, 0, 1)
        If iResult = 0 Then
            Return -1
        End If
        Return ba(0)
    End Function

End Class

